import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.beta import Beta
from torch.distributions.bernoulli import Bernoulli

class ConvNet(nn.Module):
    def __init__(self):
        super(ConvNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=5, stride=1, padding=0)
        self.conv1_mean = nn.Conv2d(3, 64, kernel_size=5, stride=1, padding=0)
        self.conv1_var = nn.Conv2d(3, 64, kernel_size=5, stride=1, padding=0)
        self.mp = nn.MaxPool2d(2)
        self.relu1 = nn.ReLU(inplace=True)
        self.relu1_1 = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(64, 128, kernel_size=5, stride=1, padding=0)
        self.conv2_mean = nn.Conv2d(64, 128, kernel_size=5, stride=1, padding=0)
        self.conv2_var = nn.Conv2d(64, 128, kernel_size=5, stride=1, padding=0)
        self.relu2 = nn.ReLU(inplace=True)
        self.relu2_1 = nn.ReLU(inplace=True)
        self.fcmix = nn.Linear(2*64*28*28, 3)
        self.sig1var = nn.Sigmoid()
        self.fc1 = nn.Linear(128*5*5, 1024)
        self.relu3 = nn.ReLU(inplace=True)
        self.fc2 = nn.Linear(1024, 1024)
        self.relu4 = nn.ReLU(inplace=True)
        self.fc3 = nn.Linear(1024, 10)

    def noise(self, mu, logvar, return_std=False):
        std = torch.exp(0.5*logvar)
        if return_std:
            return mu, std
        eps = torch.randn_like(mu)
        return mu + eps*std

    def forward(self, x, mix=True, return_feat=False, noise_layer=False, return_noise=False, return_std=False):

        in_size = x.size(0)

        if not noise_layer:
            out1 = self.relu1(self.mp(self.conv1(x)))
            out2 = self.relu2(self.mp(self.conv2(out1)))
        else:
            noise1 = F.softplus(self.noise(self.conv1_mean(x), self.conv1_var(x)))
            conv1 = self.conv1(x)
            out1 = self.relu1(self.mp(conv1))
            out1_noise = self.relu1_1(self.mp(conv1 + noise1))

            noise2 = F.softplus(self.noise(self.conv2_mean(out1_noise), self.conv2_var(out1_noise)))
            conv2 = self.conv2(out1)
            conv2_noise = self.conv2(out1_noise)
            out2_o = self.relu2(self.mp(conv2))
            out2_noise = self.relu2_1(self.mp(conv2_noise + noise2))

            conv1_mean, conv1_var = self.noise(self.conv1_mean(x), self.conv1_var(x), return_std=True)

            conv_var_cat = torch.cat([conv1_mean.reshape(in_size, -1), conv1_var.reshape(in_size, -1)], dim=1).cuda()

            absmo= self.fcmix(conv_var_cat.detach())

            a, b, smo = torch.split(absmo, 1, dim=1)
            smo = self.sig1var(smo)
            m = Bernoulli(smo)
            mask = m.sample().bool()

            if mix:
                beta = Beta(a.unsqueeze(-1).unsqueeze(-1), b.unsqueeze(-1).unsqueeze(-1))
                lam = beta.sample()
                out2 = lam * out2_o.detach() + (1 - lam) * out2_noise.detach()
            else:
                lam = 0
                out2 = out2_noise

        out2_n = out2.reshape(in_size, -1)
        out3 = self.relu3(self.fc1(out2_n))
        out4 = self.relu4(self.fc2(out3))

        if return_std:
            return self.noise(self.conv1_mean(x), self.conv1_var(x), return_std=True), \
                   self.noise(self.conv2_mean(out1_noise), self.conv2_var(out1_noise), return_std=True),\
                   a, b

        if return_noise:
            return conv1, noise1, conv2_noise, noise2

        if return_feat:
            if not noise_layer:
                return out1, out2, out4, self.fc3(out4)
            else:
                return mask, out1_noise, out2_noise, out4, self.fc3(out4)
        else:
            if noise_layer:
                return lam, mask, self.fc3(out4)
            else:
                return self.fc3(out4)