import torch.nn as nn
import torch.nn.functional as F

class LabelSmoothLoss(nn.Module):
    def __init__(self, smoothing=0.1):
        super(LabelSmoothLoss, self).__init__()
        self.smoothing = smoothing
    def forward(self, input, target, mask, lam):
        mask = mask.squeeze()
        log_prob = F.log_softmax(input, dim=-1)
        weight = input.new_zeros(input.size())
        weight_ori = input.new_zeros(input.size())
        weight[mask] += self.smoothing / (input.size(-1) - 1.)
        weight[mask] = weight[mask].scatter_(-1, target[mask].unsqueeze(-1), (1. - self.smoothing))
        weight[~mask] = weight[~mask].scatter_(-1, target[~mask].unsqueeze(-1), 1.)
        weight_ori = weight_ori.scatter_(-1, target.unsqueeze(-1), 1.)
        final_weight = lam*weight_ori + (1-lam)*weight
        loss = (-final_weight * log_prob).sum(dim=-1).mean()
        return loss